/*
    CharityWorld: SimpleAgent.m
    Copyright (C) 2003  Macaulay Institute

    This file is part of CharityWorld, a simple agent-based model to
    demonstrate the effects of errors in floating point arithmetic on
    emergent outcomes.

    CharityWorld is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    CharityWorld is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/

#import "SimpleAgent.h"
#import "Bank.h"
#import "ModelSwarm.h"
#import "MergeSorter.h"
#import "Verbosity.h"
#import "Shuffler.h"
#import "DoubleSimple.h"
#import <random.h>

@implementation SimpleAgent

-(BOOL)beggarCondition {
  DoubleSimple *superWealth = [wealth copy];
  BOOL ans;

  [self computeLocalWealth];
  [superWealth imul: (numberOfNbrs + 1)];
  ans = [superWealth lt: localWealth];

  [superWealth free];
  return ans;
}

-(BOOL)donorCondition {
  DoubleSimple *superWealth = [wealth copy];
  BOOL ans;

  [self computeLocalWealth];
  [superWealth imul: (numberOfNbrs + 1)];
  ans = [superWealth gt: localWealth];

  [superWealth free];
  return ans;
}

-(BOOL)respondingCondition: (Agent *)nbr {
  return YES;
}

-(BOOL)acceptingCondition: (Agent *)nbr {
  return YES;
}

-(BOOL)offeringCondition: (Agent *)nbr {
  DoubleSimple *superNbrWealth = [nbr getWealth];
  BOOL ans;

  [superNbrWealth imul: (numberOfNbrs + 1)];
  ans = [superNbrWealth lt: localWealth];

  [superNbrWealth free];
  return ans;
}

-(BOOL)askingCondition: (Agent *)nbr {
  DoubleSimple *superNbrWealth = [nbr getWealth];
  BOOL ans;

  [superNbrWealth imul: (numberOfNbrs + 1)];
  ans = [superNbrWealth gt: localWealth];

  [superNbrWealth free];
  return ans;
}  

-(BOOL)happinessCondition {
  DoubleSimple *superWealth = [wealth copy];
  BOOL ans;

  [self computeLocalWealth];
  [superWealth imul: (numberOfNbrs + 1)];
  ans = [superWealth gt: localWealth];

  [superWealth free];
  return ans;
}



-step {
  if([Verbosity showAgentStepping]) printf("Stepping agent (%d, %d)\n", x + 1,
					   y + 1);
  // sort the neighbours in increasing order of wealth
  [Shuffler shuffleList: nbrs];
  [MergeSorter mergeSortDouble: nbrs withCloningSelector: M(getWealth)];
  
  if([self beggarCondition]) {
    if([Verbosity showAgentSteppingDetail]) {
      printf("Agent (%d, %d) is asking because they meet "
	     "beggar condition\n", x + 1, y + 1);
    }
    [self ask];
  }
  else if([self donorCondition]) {
    if([Verbosity showAgentSteppingDetail]) {
      printf("Agent (%d, %d) is offering because they meet "
	     "donor condition\n", x + 1, y + 1);
    }
    [self offer];
  }
  else {
    if([Verbosity showAgentSteppingDetail]) {
      printf("Agent (%d, %d) is not doing anything because they"
	     " do not meet beggar or donor conditions\n", x + 1, y + 1);
    }
  }

  [self rescheduleCheck];
  return self;
}

-ask { 
  id <Index> ix = nil;
  Agent *nbr;

  if([Verbosity showAgentAsking]) {
    printf("Agent (%d, %d) is asking from their neighbours:\n", x + 1, y + 1);
  }
  for(ix = [nbrs begin: scratchZone], [ix setLoc: End],
	nbr = (Agent *)[ix prev];
      [ix getLoc] == Member;
      nbr = (Agent *)[ix prev]) {
    if([Verbosity showAgentAsking]) {
      printf("\tAgent (%d, %d) has wealth ", x + 1, y + 1);
      [wealth writeAccurately: stdout];
      printf("\n");
      printf("\tConsidering neighbour (%d, %d) with wealth ", [nbr getX] + 1,
	     [nbr getY] + 1);
      [[[nbr getWealth] writeAccurately: stdout] free];
      printf("...\n");
    }
    if([self beggarCondition]) {
      if([self askingCondition: nbr]) {
	DoubleSimple *begAmount = [beg copy];
	if([Verbosity showAgentAsking]) {
	  printf("\t...meets asking condition so telling them to respond "
		 "with ");
	  [beg writeAccurately: stdout];
	  printf("\n");
	}
        [nbr respondTo: self askingFor: begAmount];
	[begAmount free];
      }
      else {
	if([Verbosity showAgentAsking]) {
	  printf("\t...does not meet asking condition so ignoring them\n");
	}
      }
    }
    else {
      if([Verbosity showAgentAsking]) {
	printf("\t...but I no longer satisfy the beggar condition!\n");
      }
    }
  }
  [ix drop];
  for(ix = [nbrs begin: scratchZone], [ix setLoc: End],
	nbr = (Agent *)[ix prev];
      [ix getLoc] == Member;
      nbr = (Agent *)[ix prev]) {
    if(![nbr isWaiting]) {
      [nbr rescheduleCheck];
    }
  }
  [ix drop];
  return self;
}

-offer {
  id <Index> ix = nil;
  Agent *nbr;

  if([Verbosity showAgentOffering]) {
    printf("Agent (%d, %d) is offering to their neighbours:\n", x + 1, y + 1);
  }
  for(ix = [nbrs begin: scratchZone], nbr = (Agent *)[ix next];
      [ix getLoc] == Member;
      nbr = (Agent *)[ix next]) {
    if([Verbosity showAgentOffering]) {
      printf("\tConsidering neighbour (%d, %d) with wealth ", [nbr getX] + 1,
	     [nbr getY] + 1);
      [[[nbr getWealth] writeAccurately: stdout] free];
      printf("...\n");
    }
    if([self donorCondition]) {
      if([self offeringCondition: nbr]) {
	DoubleSimple *genAmount = [generosity copy];
	if([Verbosity showAgentOffering]) {
	  printf("\t...meets offering condition so telling them to accept ");
	  [generosity writeAccurately: stdout];
	  printf("\n");
	}
        [nbr accept: generosity from: self];
	[genAmount free];
      }
      else {
	if([Verbosity showAgentOffering]) {
	  printf("\t...does not meet offering condition so ignoring them\n");
	}
      }
    }
    else {
      if([Verbosity showAgentOffering]) {
	printf("\t...but I no longer satisfy the donor condition!\n");
      }
    }
  }
  [ix drop];
  for(ix = [nbrs begin: scratchZone], nbr = (Agent *)[ix next];
      [ix getLoc] == Member;
      nbr = (Agent *)[ix next]) {
    if(![nbr isWaiting]) {
      [nbr rescheduleCheck];
    }
  }
  [ix drop];
  return self;
}

-(void)accept: (DoubleSimple *)amount from: (Agent *)a {
  if([self acceptingCondition: a]) {
    if([Verbosity showAgentAccepting]) {
      printf("(%d, %d) is accepting amount ", x + 1, y + 1);
      [amount writeAccurately: stdout];
      printf(" from agent (%d, %d)\n", [a getX] + 1, [a getY] + 1);
    }
    [Bank transfer: amount from: a to: self];
  }
  else {
    if([Verbosity showAgentAccepting]) {
      printf("(%d, %d) does not accept amount ", x + 1, y + 1);
      [amount writeAccurately: stdout];
      printf(" from agent (%d, %d)\n", [a getX] + 1, [a getY] + 1);
    }
  }
}

-(void)respondTo: (Agent *)a askingFor: (DoubleSimple *)amount {
  if([self respondingCondition: a]) {
    if([Verbosity showAgentResponding]) {
      printf("(%d, %d) is responding to agent (%d, %d) asking for amount ",
	     x + 1, y + 1, [a getX] + 1, [a getY] + 1);
      [amount writeAccurately: stdout];
      printf("\n");
    }
    [Bank transfer: amount from: self to: a];
  }
  else {
    if([Verbosity showAgentResponding]) {
      printf("(%d, %d) does not respond to agent (%d, %d) asking for amount ",
	     x + 1, y + 1, [a getX] + 1, [a getY] + 1);
      [amount writeAccurately: stdout];
      printf("\n");
    }
  }
}

-(void)reschedule {
  waiting = YES;
  if([self beggarCondition]) {
    if([Verbosity showAgentRescheduling]) {
      printf("(%d, %d) queued to beggar list\n", x + 1, y + 1);
    }
    [model queueBeggar: self];
  }
  else if([self donorCondition]) {
    if([Verbosity showAgentRescheduling]) {
      printf("(%d, %d) queued to donor list\n", x + 1, y + 1);
    }
    [model queueDonor: self];
  }
  else {
    if([Verbosity showAgentRescheduling]) {
      printf("(%d, %d) not queued to any list\n", x + 1, y + 1);
    }
    waiting = NO;
  }
}  

@end
